"""ROC curves for classification models."""

import logging
from typing import Callable

import plotly.graph_objs as go
from sklearn.metrics import roc_curve

from ml_inspector._metrics_curves import MetricsCurves

logger = logging.getLogger(__name__).setLevel(logging.INFO)


class ROCCurves(MetricsCurves):
    """A class to display ROC curves."""

    def __init__(
        self,
        curve_name: str = "ROC",
        curve_function: Callable = roc_curve,
        xaxis: dict = {"index": 0, "name": "False Positive Rate", "range": (0, 1)},
        yaxis: dict = {"index": 1, "name": "True Positive Rate", "range": (0, 1)},
    ) -> None:
        super().__init__(curve_name, curve_function, xaxis, yaxis)

    def add_random_decision(self, *args, **kwargs) -> go.Scatter:
        """Returns the metrics values for a random decision."""
        fig = go.Scatter(
            x=[0, 1],
            y=[0, 1],
            line={"color": "black", "dash": "dash"},
            name="Random decision: AUC=0.50",
            mode="lines",
        )
        return fig


def plot_roc_curves(y_true, y_prob, class_names=None, decision_threshold=None):
    """Plots the ROC curves for a binary or multi-class classification model.

    :param array y_true:
        An array containing the true outcomes.
    :param dict y_prob:
        A dictionary containing the predicted probablities for each class, together
        with their labels. For example:
            {"Train": array([[0.3, 0.7], ...]), "Test": array([[0.4, 0.6], ...])}
        In case of binary classification, only the probablities
        for the positive class may be provided. For example:
            {"Model 1": array([0.1, 0.4]), "Model 2": array([0.2, 0.3])}
        Alternatively, a single array may be provided. For example:
            array([[0.1, 0.9], [0.2, 0.8], ...])
    :param dict class_names:
        A dictionary containing the name to display for each class. For example:
            {0: "Class 0", 1: "Class 1", ...}).
    :param dict decision_threshold:
        A dictionary containing the threshold indicating where the boolean decision is
        made from probablity predictions (for binary classification models) together
        with their labels. For example:
            {"Train": 0.5, "Test": 0.7}

    :returns plotly.graph_objs.Figure:
        The figure containing the ROC curves.
    """
    fig = ROCCurves().plot_curves(y_true, y_prob, class_names, decision_threshold)
    return fig



from .calibration_curves import plot_calibration_curves
from .confusion_matrix import plot_confusion_matrix
from .feature_importance import plot_feature_importance
from .gain_curves import plot_gain_curves
from .learning_curves import plot_learning_curves
from .partial_dependence import plot_partial_dependence
from .precision_recall_curves import plot_precision_recall_curves
from .predictions import plot_classification_predictions
from .roc_curves import plot_roc_curves

__all__ = [
    "plot_calibration_curves",
    "plot_confusion_matrix",
    "plot_gain_curves",
    "plot_learning_curves",
    "plot_partial_dependence",
    "plot_precision_recall_curves",
    "plot_classification_predictions",
    "plot_roc_curves",
    "plot_feature_importance",
]



"""Gain curves for classification models."""

import logging
from typing import Callable

import numpy as np
import pandas as pd
import plotly.graph_objs as go

from ml_inspector._metrics_curves import MetricsCurves

logger = logging.getLogger(__name__).setLevel(logging.INFO)


def calculate_gain_curve(y_true: np.ndarray, y_prob: np.ndarray) -> tuple:
    """Calculate the gain curve as a function of the threshold value.

    :param array y_true:
        An array containing the true binary outcomes (0s and 1s).
    :param array y_prob:
        An array containing the predicted probablities for the positive class.

    :returns tuple:
        A tuple containing:
        * the fractions of samples
        * the fractions of detected positive class samples
        * the corresponding thresholds
    """
    y_prob = pd.Series(np.array(y_prob)).sort_values(ascending=False)
    y_true = pd.Series(np.array(y_true)).reindex_like(y_prob)
    recalls = y_true.cumsum() / y_true.sum()
    fractions = [i / len(y_true) for i in range(len(y_true))]
    thresholds = y_prob
    return np.array(fractions), np.array(recalls), np.array(thresholds)


class GainCurves(MetricsCurves):
    """A class to display gain curves."""

    def __init__(
        self,
        curve_name: str = "Gain",
        curve_function: Callable = calculate_gain_curve,
        xaxis: dict = {"index": 0, "name": "Fraction selected", "range": (0, 1)},
        yaxis: dict = {"index": 1, "name": "Fraction detected", "range": (0, 1)},
    ) -> None:
        super().__init__(curve_name, curve_function, xaxis, yaxis)

    def add_random_decision(self, *args, **kwargs) -> go.Scatter:
        """Returns the metrics values for a random decision."""
        fig = go.Scatter(
            x=[0, 1],
            y=[0, 1],
            line={"color": "black", "dash": "dash"},
            name="Random decision: AUC=0.50",
            mode="lines",
        )
        return fig


def plot_gain_curves(y_true, y_prob, class_names=None, decision_threshold=None):
    """Plots the gain curves for a binary or multi-class classification model.

    :param array y_true:
        An array containing the true outcomes.
    :param dict y_prob:
        A dictionary containing the predicted probablities for each class, together
        with their labels. For example:
            {"Train": array([[0.3, 0.7], ...]), "Test": array([[0.4, 0.6], ...])}
        In case of binary classification, only the probablities
        for the positive class may be provided. For example:
            {"Model 1": array([0.1, 0.4]), "Model 2": array([0.2, 0.3])}
        Alternatively, a single array may be provided. For example:
            array([[0.1, 0.9], [0.2, 0.8], ...])
    :param dict class_names:
        A dictionary containing the name to display for each class. For example:
            {0: "Class 0", 1: "Class 1", ...}).
    :param dict decision_threshold:
        A dictionary containing the threshold indicating where the boolean decision is
        made from probablity predictions (for binary classification models) together
        with their labels. For example:
            {"Train": 0.5, "Test": 0.7}

    :returns plotly.graph_objs.Figure:
        The figure containing the gain curves.
    """
    fig = GainCurves().plot_curves(y_true, y_prob, class_names, decision_threshold)
    return fig



"""Functions to display partial depencence plots."""

import numpy as np
import pandas as pd
from plotly import graph_objs as go
from plotly.colors import DEFAULT_PLOTLY_COLORS
from tqdm.auto import tqdm


def select_feature_values(X, feature, max_nb_points):
    """Returns a list of values for the selected feature. If the feature is numerical
    then its quantiles are used as the feature values. If the feature is a string
    (categorical feature), the most frequent values are returned.

    :param pandas.DataFrame X:
        A pandas DataFrame containing the training data.
    :param str feature:
        The feature for which to return the values.
    :param int max_nb_points:
        The maximum number of values returned for the feature.

    :returns list:
        The list of values for the selected feature.
    """
    if X[feature].dtype == "object":
        feature_values = list(X[feature].value_counts().index)[:max_nb_points]
    else:
        X = X.copy()
        X[feature] = remove_outliers(X[feature], sigma_factor=0)
        feature_values = [
            X[feature].quantile(i / (max_nb_points - 1), interpolation="nearest")
            for i in range(max_nb_points)
        ]
        feature_values = sorted(set(feature_values))
    return feature_values


def ceteris_paribus(estimator, row, feature, feature_values, **kwargs):
    """Returns the estimator probability predictions as a function of the values
    for a selected feature for a single row of data (all other features remaining
    equal).

    :type estimator: A fitted sklearn estimator.
    :param estimator:
        The estimator with which to make predictions.
    :param pandas.Series row:
        A pandas Series containing the features for the selected row.
    :param str feature:
        The feature for which to estimate the ceteris paribus dependence.
    :param list feature_values:
        The list of values for the feature with which to make predictions with.

    :returns numpy.array:
        The list of ceteris paribus predictions for the row of data for each
        of the values for the selected feature.
    """
    X_ceteris_paribus = pd.DataFrame([row] * len(feature_values))
    X_ceteris_paribus[feature] = feature_values
    if estimator._estimator_type == "classifier":
        predictions = estimator.predict_proba(X_ceteris_paribus)
        if len(estimator.classes_) == 2:
            predictions = predictions[:, 1]
        else:
            predictions = np.array(
                [predictions[:, i] for i in range(len(estimator.classes_))]
            )
    else:
        predictions = estimator.predict(X_ceteris_paribus, **kwargs)
    return predictions


def partial_dependence(
    estimator, X, feature, max_nb_points=20, max_sample=50, **kwargs
):
    """Returns all the estimator predictions for a sample of rows drawn from
    the training dataset by only varying the selected feature over the
    range of values.

    :type estimator: A fitted sklearn estimator.
    :param estimator:
        The estimator for which to assess the partial dependence on a feature.
    :param pandas.DataFrame X:
        A pandas DataFrame containing the training data.
    :param str feature:
        The feature for which to assess the partial dependence of the estimator.
    :param int max_nb_points:
        The maximum number feature values.
    :param int max_sample:
        The number of samples from the training data on which to estimate the
        partial dependence of the model on the feature.

    :returns tuple:
        A tuple containing the values that the selected feature may take as well
        as the ceteris paribus predictions for these feature values, and the
        impact of each predictions relative to the average predictions.
    """
    all_predictions, all_impacts = [], []
    if len(X) < max_sample:
        max_sample = len(X)
    X_sample = X.sample(max_sample)
    feature_values = select_feature_values(X, feature, max_nb_points)
    for index, row in X_sample.iterrows():
        predictions = ceteris_paribus(estimator, row, feature, feature_values, **kwargs)
        all_predictions.append(predictions)
        if predictions.ndim == 1:
            impacts = predictions - np.mean(predictions)
        elif predictions.ndim == 2:
            impacts = np.subtract(predictions.T, np.mean(predictions, axis=1)).T
        all_impacts.append(impacts)
    return feature_values, np.array(all_predictions), np.array(all_impacts)


def plot_partial_dependence(
    estimator,
    X,
    features=None,
    max_nb_points=20,
    max_sample=100,
    class_names=None,
    **kwargs,
):
    """Plots the partial dependence of an estimator on a selected feature as well
    as the confidence interval for the dependence based on a sample of rows
    drawn from the training data.

    :type estimator: A fitted sklearn estimator.
    :param estimator:
        The estimator for which to assess the partial dependence on a feature.
    :param pandas.DataFrame X:
        A pandas DataFrame containing the training data.
    :param list features:
        A list of features for which to assess the partial dependence of the model.
    :param int max_nb_points:
        The maximum number feature values.
    :param int max_sample:
        The number of samples from the training data on which to estimate the
        partial dependence of the model on the feature.
    :param dict class_names:
        A dictionary containing the name to display for each class
        (e.g. {1: "Class 1", 2: "Class 2", ...}).
    :param str save_path:
        The path to export the figure to.
    :param bool display:
        A flag to display the partial dependence plot.
    """
    plot_data = []
    visible = True
    if features is not None:
        features = X.columns
    for feature in tqdm(features, desc="Calculating partial dependence"):
        feature_values, all_predictions, all_impacts = partial_dependence(
            estimator,
            X,
            feature,
            max_nb_points=max_nb_points,
            max_sample=max_sample,
            **kwargs,
        )
        if all_predictions[0].ndim == 2:
            if not class_names:
                class_names = {c: f"Class {c}" for c in estimator.classes_}
            class_names = {i: class_names[i] for i in estimator.classes_}
        plot_data.extend(
            partial_dependence_plot_data(
                feature_values, all_impacts, class_names, visible=visible
            )
        )
        visible = False
    layout = partial_dependence_plot_layout(X.columns[0])
    fig = go.Figure(data=plot_data, layout=layout)
    fig = add_feature_selection_button(fig, X, max_sample)
    return fig


def partial_dependence_plot_data(
    feature_values, all_impacts, class_names=None, ci=95, visible=True
):
    """Returns the data for the partial dependence plot and its confidence interval
    as well as the individual ceteris paribus plots.

    :param list feature_values:
        The list of feature values for the partial dependence plot.
    :param list all_predictions:
        A list of arrays containing the predictions for each feature value and
        for each row.
    :param list all_impacts:
        A list of arrays containing the impact of the feature relative to the
        average prediction for each feature value and for each row.
    :param dict class_names:
        A dictionary containing the name to display for each class
        (e.g. {1: "Class 1", 2: "Class 2", ...}).
    :param float ci:
        The confidence interval to display for the partial dependence plot (in
        percents).
    :param bool visible:
        A flag to display the individual ceteris paribus plots.

    :returns list:
        A list of plotly traces containing the partial dependence plot data.
    """
    data = []
    data.extend(add_individual_impact_plots(feature_values, all_impacts, visible))
    data.extend(
        add_average_impact_plots(feature_values, all_impacts, class_names, ci, visible)
    )
    return data


def add_individual_impact_plots(feature_values, all_impacts, visible):
    """Generates the individual sample partial dependence plots.

    :param list feature_values:
        The list of feature values for the partial dependence plot.
    :param list all_impacts:
        A list of arrays containing the impact of the feature relative to the
        average prediction for each feature value and for each row.

    :returns list:
        A list of plotly traces containing the partial dependence plot data.
    """
    data = []
    if all_impacts[0].ndim == 1:
        for i, impact in enumerate(all_impacts):
            data.append(
                go.Scatter(
                    x=feature_values,
                    y=impact,
                    mode="lines+markers",
                    marker={"color": "grey", "size": 1},
                    name="Individual predictions",
                    showlegend=(i == 0),
                    legendgroup="ceteris_paribus",
                    opacity=0.5,
                    hoverinfo="skip",
                    visible=visible,
                )
            )
    return data


def add_average_impact_plots(feature_values, all_impacts, class_names, ci, visible):
    """Add the average partial dependence and confidence interval plots. If there is
    more than two classes, one plot per class is created.

    :param list feature_values:
        The list of feature values for the partial dependence plot.
    :param list all_predictions:
        A list of arrays containing the predictions for each feature value and
        for each row.
    :param dict class_names:
        A dictionary containing the name to display for each class
        (e.g. {1: "Class 1", 2: "Class 2", ...}).
    :param float ci:
        The confidence interval to display for the partial dependence plot (in
        percents).

    :returns list:
        A list of plotly traces containing the partial dependence plot data.
    """
    data = []
    if all_impacts[0].ndim == 1:
        all_impacts = np.array([[x] for x in all_impacts])
    nb_plots = len(all_impacts[0])
    for i in range(nb_plots):
        mean_values = np.mean(all_impacts[:, i], axis=0)
        pos_std = np.percentile(all_impacts[:, i], q=100 - (100 - ci) / 2, axis=0)
        neg_std = np.percentile(all_impacts[:, i], q=(100 - ci) / 2, axis=0)
        color = DEFAULT_PLOTLY_COLORS[i]
        name = f" ({list(class_names.values())[i]})" if class_names else ""
        data.extend(
            [
                go.Scatter(
                    x=feature_values,
                    y=pos_std,
                    mode="lines+markers",
                    marker={"color": color, "size": 1},
                    name=f"{ci}% Confidence Interval" + f"{name}",
                    line={"width": 2, "dash": "dash"},
                    opacity=0.5,
                    legendgroup=i,
                    showlegend=False,
                    visible=visible,
                ),
                go.Scatter(
                    x=feature_values,
                    y=neg_std,
                    mode="lines+markers",
                    marker={"color": color, "size": 1},
                    name=f"{ci}% Confidence Interval" + f"{name}",
                    line={"width": 2, "dash": "dash"},
                    fill="tonexty",
                    opacity=0.5,
                    legendgroup=i,
                    visible=visible,
                ),
                go.Scatter(
                    x=feature_values,
                    y=mean_values,
                    mode="lines+markers",
                    marker={"color": color, "size": 6},
                    line={"width": 5},
                    name="Average partial dependence" + f"{name}",
                    legendgroup=i,
                    visible=visible,
                ),
            ]
        )
    return data


def partial_dependence_plot_layout(feature):
    """Returns the layout for the partial dependence plot.

    :param str feature:
        The feature for which to create the partial dependence plot.

    :returns go.Layout:
        The layout for the partial dependence plot.
    """
    layout = go.Layout(
        template="none",
        xaxis={"title": f"Value of {feature}"},
        yaxis={"title": "Impact on model prediction", "tickformat": ".2%"},
        height=600,
        width=1000,
        legend={"orientation": "h", "y": 1.1},
    )
    return layout


def remove_outliers(series, q_min=0.01, q_max=0.99, sigma_factor=0):
    """Removes the outliers from a pandas Series by selecting the values
    between a minimum and maximum quantiles to which can be added a
    number of standard deviations.

    :param pandas.Series series:
        The pandas Series for which to remove the outliers.
    :param float q_min:
        The lower quantile below which to filter values.
    :param float q_max:
        The higher quantile above which to filter values.
    :param float sigma_factor:
        The number of standard deviations to include above and below the higher
        and lower quantiles when filtering the data.

    :returns pandas.Series
        The pandas Series without its outliers.
    """
    lower_limit = series.quantile(q_min) - sigma_factor * series.std()
    upper_limit = series.quantile(q_max) + sigma_factor * series.std()
    return series[series.between(lower_limit, upper_limit)].copy()


def add_feature_selection_button(fig, X, max_sample):
    repeat = max_sample + 3
    buttons = [
        {
            "args": [
                {"visible": [c == col for c in X.columns for _ in range(repeat)]},
                {"title": col, "xaxis": {"title": f"Value of {col}"}},
            ],
            "label": col,
            "method": "update",
        }
        for col in X.columns
    ]
    fig.update_layout(
        updatemenus=[
            dict(
                buttons=buttons,
                direction="down",
                pad={"r": 10, "t": 10},
                showactive=True,
                x=0,
                xanchor="left",
                y=1.3,
                yanchor="top",
            ),
        ]
    )
    return fig



from typing import Callable

import numpy as np
import plotly.graph_objects as go
from plotly.colors import DEFAULT_PLOTLY_COLORS
from sklearn.metrics import auc
from sklearn.preprocessing import label_binarize


class MetricsCurves:
    """A class to display metrics curves for classification models.

    Args:
        curve_name (str): The name of the metrics to be displayed.
        curve_function (Callable): The function to calculate the metrics curve.
        xaxis (dict): A dictionary containing the x-axis information, including the
            index in the curve data, the name and the range.
        yaxis (dict): A dictionary containing the y-axis information, including the
            index in the curve data, the name and the range.
    """

    def __init__(
        self,
        curve_name: str,
        curve_function: Callable,
        xaxis: dict,
        yaxis: dict,
        reverse_thresholds: bool = False,
    ) -> None:
        self.curve_name = curve_name
        self.curve_function = curve_function
        self.xaxis = xaxis
        self.yaxis = yaxis
        self.reverse_thresholds = reverse_thresholds

    def plot_curves(
        self,
        y_true: np.ndarray,
        y_prob: dict[str, np.ndarray] | np.ndarray,
        class_names: dict | None = None,
        decision_threshold: float | None = None,
    ) -> go.Figure:
        """Plots the metrics curves based on the ground truth and predictions.

        Args:
            y_true (array): An array containing the true outcomes.
            y_prob (dict): A dictionary containing the predicted probablities for
                each class, together with their labels. For example:
                {"Train": [[0.3, 0.7], ...], "Test": [[0.4, 0.6], ...]}
                In case of binary classification, only the probablities
                for the positive class may be provided. For example:
                {"Model 1": [0.1, 0.4, ...], "Model 2": [0.2, 0.3, ...]}
                Alternatively, a single array may be provided. For example:
                [[0.1, 0.9], [0.2, 0.8], ...]
            class_names (dict): A dictionary containing the name to display
                for each class. For example: {0: "Class 0", 1: "Class 1", ...}).
            decision_threshold (float): The probablity above which the class
                is predicted (for binary classification models only).
        """
        y_prob = {"Predictions": y_prob} if isinstance(y_prob, np.ndarray) else y_prob
        classes = np.unique(y_true)
        if len(classes) < 2:
            raise ValueError("Metrics curves are not defined for less than two classes")
        curves_data = self.calculate_curves_data(y_true, y_prob, classes)
        if class_names is None:
            class_names = {c: str(c) for c in classes}
        fig = self.create_curves_plot(curves_data, class_names, decision_threshold)
        return fig

    def calculate_curves_data(
        self, y_true: np.ndarray, y_prob: dict[str, np.ndarray], classes: list
    ) -> dict[str, tuple]:
        """Calculates the metrics curves.

        Args:
            y_true (array): An array containing the true outcomes.
            y_prob (dict): A dictionary containing the predicted probablities for
                each class.
            classes (list): A list containing the class labels.

        Returns:
            dict: The metrics curves for each class.
        """
        curves_data = {}
        for k, values in y_prob.items():
            if len(classes) == 2:
                curves_data[k] = self.get_binary_curve(y_true, values)
            else:
                curves_data[k] = self.get_multiclass_curve(y_true, values, classes)
        return curves_data

    def get_binary_curve(self, y_true: np.ndarray, y_prob: np.ndarray) -> dict:
        """Calculates the metrics curve for binary classification models.

        Args:
            y_true (array): An array containing the true outcomes.
            y_prob (array): An array containing the predicted probablities. If y_prob
                contains probablities for both classes, the curve is calculated for
                the positive class.

        Returns:
            dict: The metrics curve data for the positive class.
        """
        curve_data = {}
        if y_prob.shape[1] == 2:
            y_prob = y_prob[:, 1]
        curve_data[1] = self.curve_function(y_true, y_prob)
        return curve_data

    def get_multiclass_curve(
        self, y_true: np.ndarray, y_prob: np.ndarray, classes: list
    ) -> dict:
        """Calculates the metrics curve for multi-class classification models for each
        class as well as for the micro-average.

        Args:
            y_true (array): An array containing the true outcomes.
            y_prob (array): An array containing the predicted probablities for each
                class.
            classes (list): A list containing the class labels.

        Returns:
            dict: The metrics curve data for each class and the micro-average.
        """
        curve_data = {}
        for i, c in enumerate(classes):
            curve_data[c] = self.curve_function(y_true == c, y_prob[:, i])
        curve_data["Average"] = self.curve_function(
            label_binarize(y_true, classes=classes).ravel(), y_prob.ravel()
        )
        return curve_data

    def create_curves_plot(self, curves: dict, class_names: list, threshold: float):
        """Generates a plotly figure of the metrics curves.

        Args:
            curves (dict): A dictionary containing the metrics curves to display.
            class_names (dict): A dictionary containing the name to display for each
                class. For example: {1: "Class 1", 2: "Class 2", ...}.
            threshold (float): The probablity above which the class is predicted
                (for binary classification models only).

        Returns:
            go.Figure: The figure containing the metrics curves.
        """
        data = self.create_curves_plot_data(curves, class_names, threshold)
        layout = self.create_curves_plot_layout()
        fig = go.Figure(data=data, layout=layout)
        return fig

    def create_curves_plot_data(
        self, curves: dict, class_names: dict, threshold: float
    ):
        """Creates the data for the metrics curves plot.

        Args:
            curves (dict): A dictionary containing the metrics curves to display.
            class_names (dict): A dictionary containing the name to display for each
                class. For example: {1: "Class 1", 2: "Class 2", ...}.
            threshold (float): The probablity above which the class is predicted
                (for binary classification models only).

        Returns:
            list: A list of scatter plots for the metrics curves.
        """
        class_names = class_names.copy()
        if len(class_names) > 2:
            class_names["Average"] = f"Micro-average {self.curve_name} curve"
        lines = ["solid", "dot", "dash", "longdash", "dashdot", "longdashdot"]
        data = []
        for j, type_curves in enumerate(curves):
            for i, c in enumerate(curves[type_curves]):
                curve = curves[type_curves][c]
                color = DEFAULT_PLOTLY_COLORS[i]
                if threshold:
                    showlegend = j == 0 and i == 0
                    data.append(
                        self.plot_threshold(
                            curve, threshold, color, showlegend, legendgroup=str(c)
                        )
                    )
                x_vals = curve[self.xaxis["index"]]
                y_vals = curve[self.yaxis["index"]]
                score = auc(x_vals, y_vals)
                name = f"{class_names[c]} ({type_curves}): AUC={score:.2f}"
                line = {"color": color, "dash": lines[j % len(lines)]}
                data.append(
                    go.Scatter(
                        x=x_vals, y=y_vals, line=line, name=name, legendgroup=str(c)
                    )
                )
        if hasattr(self, "add_random_decision"):
            data.append(self.add_random_decision(x_vals=x_vals, y_vals=y_vals))
        return data

    def plot_threshold(self, curve, threshold, color, showlegend, legendgroup):
        """Plots the decision threshold for binary classification models.

         Args:
            curve (tuple): The curve data as a function of the threshold value.
            threshold (float): The selected decision threshold for the model.
            color (str): The RGB color to display the decision threshold.
            showlegend (bool): A flag to show the legend for the decision threshold.
            legendgroup (str): The legend group to attach the decision threshold to.

        Returns:
            go.Scatter: A scatter plot for the decision threshold.
        """
        x_vals = curve[self.xaxis["index"]]
        y_vals = curve[self.yaxis["index"]]
        thresholds = curve[2]
        index = len([i for i in thresholds if i >= threshold])
        if self.reverse_thresholds:
            index = len(thresholds) - index
        y = y_vals[[index - 1]]
        x = x_vals[[index - 1]]
        params = {
            "line": {"color": "black", "width": 2, "dash": "dot"},
            "name": f"Decision threshold ({threshold:.1%})",
            "legendgroup": legendgroup,
            "mode": "markers",
            "marker": {"color": color, "size": 8},
        }
        return go.Scatter(x=x, y=y, showlegend=showlegend, **params)

    def create_curves_plot_layout(self):
        """Creates the layout for the metrics curves plot.

        Returns:
            go.Layout: The layout for the metrics curves plot.
        """
        layout = go.Layout(
            legend={"traceorder": "grouped", "orientation": "h", "y": -0.1},
            width=800,
            height=800,
            xaxis={"title_text": self.xaxis["name"], "range": self.xaxis["range"]},
            yaxis={"title_text": self.yaxis["name"], "range": self.yaxis["range"]},
        )
        return layout



"""Functions to display machine learning model predictions."""

import numpy as np
from plotly import graph_objs as go
from plotly.colors import DEFAULT_PLOTLY_COLORS


def plot_classification_predictions(
    y_true: np.ndarray,
    y_prob: np.ndarray,
    class_names: dict | None = None,
    decision_threshold: float | None = None,
    points: bool | str = False,
):
    """Display the distribution of classification model predictions using violinplots.
    For binary classification, only the predicted probability distributions for
    the positive class is provided. For multiple classification predictions, the
    predicted probability distributions are provided for each combination of
    true and predicted classes (N x N distributions, where N is the number of
    classes).

    Args:
        y_true (array): An array containing the true class for each prediction.
            For example: [0, 1, 1]
        y_prob (array): An array containing the predicted probability for each class
            and for each prediction. For example: [[0.9, 0.1], [0.6, 0.4], ...]
        class_names (dict): A dictionary containing the name to display for each class.
            For example: {0: "Class 0", 1: "Class 1", ...}).
        decision_threshold (float): The probablity above which the class is predicted
            (for binary classification models only).
        points (bool or str): A flag to display the individual predicted probabilities.

    Returns:
        go.Figure: The figure containing the classification predictions distribution.
    """
    classes = np.unique(y_true)
    if not class_names:
        class_names = {c: str(c) for c in sorted(classes)}
    if len(class_names) == 2:
        plots_data = binary_distribution(y_true, y_prob, class_names, points)
    else:
        plots_data = multiclass_distribution(y_true, y_prob, class_names, points)
    layout = create_classification_distribution_layout(class_names)
    fig = go.Figure(data=plots_data, layout=layout)
    if len(class_names) == 2:
        fig = add_decision_threshold(fig, decision_threshold)
    return fig


def binary_distribution(y_true, y_pred, class_names, points):
    """Returns violin plots showing the predicted probability for the positive class
    considering the true label.

    Args:
        y_true (array): An array containing the true class for each prediction.
        y_pred (array): An array containing the predicted probability for each class
            and for each prediction.
        class_names (dict): A dictionary containing the name to display for each class.
            For example: {0: "Class 0", 1: "Class 1", ...}).
        points (bool or str): A flag to display the individual predicted probabilities.

    Returns:
        list: The violin plots data.
    """
    positive_class = max(class_names)
    plots_data = []
    for i, true_class in enumerate(class_names):
        y_class = [
            pc[positive_class] for tc, pc in zip(y_true, y_pred) if tc == true_class
        ]
        plots_data.append(
            go.Violin(
                y=y_class,
                points=points,
                pointpos=0,
                jitter=1,
                marker={"opacity": 0.2, "color": DEFAULT_PLOTLY_COLORS[i]},
                name=class_names[true_class],
            )
        )
    return plots_data


def multiclass_distribution(y_true, y_pred, class_names, points):
    """Returns violin plots showing the predicted probability for each class
    considering the true label.

    Args:
        y_true (array): An array containing the true class for each prediction.
        y_pred (array): An array containing the predicted probability for each class
            and for each prediction.
        class_names (dict): A dictionary containing the name to display for each class.
            For example: {0: "Class 0", 1: "Class 1", ...}).
        points (bool or str): A flag to display the individual predicted probabilities.

    Returns:
        list: The violin plots data.
    """
    plots_data = []
    for i, pred_class in enumerate(class_names):
        for j, true_class in enumerate(class_names):
            y_class = [pc[i] for tc, pc in zip(y_true, y_pred) if tc == true_class]
            plots_data.append(
                go.Violin(
                    y=y_class,
                    points=points,
                    pointpos=0,
                    jitter=1,
                    marker={"opacity": 0.2, "color": DEFAULT_PLOTLY_COLORS[j]},
                    yaxis=f"y{i+1}",
                    name=class_names[true_class],
                    showlegend=(i == 0),
                    legendgroup=j,
                )
            )
    return plots_data


def create_classification_distribution_layout(class_names):
    """Returns the layout for the classification predictions distribution plot.

    Args:
        class_names (dict): A dictionary containing the name to display for each class.

    Returns:
        go.Layout: The layout for the classification predictions distribution plot.
    """
    yaxes = {}
    if len(class_names) == 2:
        yaxes["yaxis"] = {
            "range": [0, 1],
            "title": f"P({class_names[max(class_names)]})",
        }
    else:
        for i, name in enumerate(class_names.values()):
            domain = [1 - (i + 1) / len(class_names), 1 - i / len(class_names) - 0.02]
            yaxes[f"yaxis{i+1}"] = {
                "domain": domain,
                "range": [0, 1],
                "title": f"P({name})",
            }
    layout = go.Layout(
        xaxis={"title": "Actual outcome", "anchor": "y"},
        width=800,
        height=250 * len(class_names),
        legend={"orientation": "h"},
        **yaxes,
    )
    return layout


def add_decision_threshold(fig, decision_threshold):
    """Adds the decision threshold to the figure (for binary classification only).

    Args:
        fig (go.Figure): The plotly figure to which to add the decision threshold.
        decision_threshold (float): The probablity above which the positive class is
            predicted.

    Returns:
        go.Figure: The plotly figure with the decision threshold.
    """
    if decision_threshold:
        fig.add_hline(
            decision_threshold,
            line_width=3,
            line_dash="dash",
            line_color="black",
            annotation_text=f"Threshold: {decision_threshold:.1%}",
            annotation_position="top left",
        )
    return fig



"""Precision-Recall curves for classification models."""

import logging
from typing import Callable

import plotly.graph_objs as go
from sklearn.metrics import precision_recall_curve

from ml_inspector._metrics_curves import MetricsCurves

logger = logging.getLogger(__name__).setLevel(logging.INFO)


class PRCurves(MetricsCurves):

    def __init__(
        self,
        curve_name: str = "Precision-Recall",
        curve_function: Callable = precision_recall_curve,
        xaxis: dict = {"index": 1, "name": "Recall", "range": (0, 1)},
        yaxis: dict = {"index": 0, "name": "Precision", "range": (0, 1)},
        reverse_thresholds: bool = True,
    ) -> None:
        super().__init__(curve_name, curve_function, xaxis, yaxis, reverse_thresholds)

    def add_random_decision(self, y_vals, *args, **kwargs) -> go.Scatter:
        """Returns the metrics values for a random decision.

        Args:
        """
        fig = go.Scatter(
            x=[0, 1],
            y=[min(y_vals), min(y_vals)],
            line={"color": "black", "dash": "dash"},
            name=f"Random decision: AUC={min(y_vals):.2f}",
            mode="lines",
        )
        return fig


def plot_precision_recall_curves(
    y_true, y_prob, class_names=None, decision_threshold=None
):
    """Plots the Precision-Recall curves for a binary or multi-class classification
    model.

    :param array y_true:
        An array containing the true outcomes.
    :param dict y_prob:
        A dictionary containing the predicted probablities for each class, together
        with their labels. For example:
            {"Train": array([[0.3, 0.7], ...]), "Test": array([[0.4, 0.6], ...])}
        In case of binary classification, only the probablities
        for the positive class may be provided. For example:
            {"Model 1": array([0.1, 0.4, ...]), "Model 2": array([0.2, 0.3, ...])}
    :param dict class_names:
        A dictionary containing the name to display for each class. For example:
            {0: "Class 0", 1: "Class 1", ...}).
    :param dict decision_threshold:
        A dictionary containing the threshold indicating where the boolean decision is
        made from probablity predictions (for binary classification models) together
        with their labels. For example:
            {"Train": 0.5, "Test": 0.7}

    :returns plotly.graph_objs.Figure:
        The figure containing the Precision-Recall curves.
    """
    fig = PRCurves().plot_curves(y_true, y_prob, class_names, decision_threshold)
    return fig



"""Tests for ml_inspector.roc_curve"""

import numpy as np
import pytest
from plotly import graph_objs as go

from ml_inspector import plot_roc_curves


class TestROCCurve:
    def test_plot_roc_curves_binary(self, binary_predictions):
        y, y_prob_1, y_prob_2 = binary_predictions
        y_prob = {"Model 1": y_prob_1, "Model 2": y_prob_2}
        class_names = {0: "Class 0", 1: "Class 1"}
        fig = plot_roc_curves(y, y_prob, class_names=class_names)
        assert isinstance(fig, go.Figure)
        assert len(fig.data) == 3
        assert fig.data[0]["name"] == "Class 1 (Model 1): AUC=1.00"
        assert fig.data[1]["name"] == "Class 1 (Model 2): AUC=0.75"
        assert fig.data[2]["name"] == "Random decision: AUC=0.50"

    def test_plot_roc_curves_multi_class(self, multiclass_predictions):
        y, y_prob_1, y_prob_2 = multiclass_predictions
        y_prob = {"Training": y_prob_1, "Test": y_prob_2}
        class_names = {0: "Class 0", 1: "Class 1", 2: "Class 2", 3: "Class 3"}
        fig = plot_roc_curves(y, y_prob, class_names=class_names)
        assert isinstance(fig, go.Figure)
        assert len(fig.data) == 11
        assert fig.data[0]["name"] == "Class 0 (Training): AUC=1.00"
        assert fig.data[1]["name"] == "Class 1 (Training): AUC=1.00"
        assert fig.data[2]["name"] == "Class 2 (Training): AUC=1.00"
        assert fig.data[3]["name"] == "Class 3 (Training): AUC=1.00"
        assert fig.data[4]["name"] == "Micro-average ROC curve (Training): AUC=0.96"
        assert fig.data[5]["name"] == "Class 0 (Test): AUC=0.88"
        assert fig.data[-1]["name"] == "Random decision: AUC=0.50"

    def test_plot_roc_curves_error_single_class(self, binary_predictions):
        y, y_prob_1, y_prob_2 = binary_predictions
        y = np.array([0, 0, 0, 0, 0])
        y_prob = {"Model 1": y_prob_1, "Model 2": y_prob_2}
        with pytest.raises(ValueError) as ve:
            plot_roc_curves(y, y_prob)
        assert str(ve.value) == (
            "Metrics curves are not defined for less than two classes"
        )

    def test_plot_roc_curves_with_threshold(self, binary_predictions):
        y, y_prob_1, y_prob_2 = binary_predictions
        y_prob = {"Model 1": y_prob_1, "Model 2": y_prob_2}
        class_names = {0: "Class 0", 1: "Class 1"}
        fig = plot_roc_curves(
            y, y_prob, class_names=class_names, decision_threshold=0.4
        )
        assert isinstance(fig, go.Figure)
        assert len(fig.data) == 5
        assert fig.data[0]["name"] == "Decision threshold (40.0%)"
        assert fig.data[1]["name"] == "Class 1 (Model 1): AUC=1.00"
        assert fig.data[2]["name"] == "Decision threshold (40.0%)"
        assert fig.data[3]["name"] == "Class 1 (Model 2): AUC=0.75"
        assert fig.data[4]["name"] == "Random decision: AUC=0.50"

    def test_plot_roc_curves_with_array(self, binary_predictions):
        y, y_prob_1, _ = binary_predictions
        class_names = {0: "Class 0", 1: "Class 1"}
        fig = plot_roc_curves(y, y_prob_1, class_names=class_names)
        assert isinstance(fig, go.Figure)
        assert len(fig.data) == 2
        assert fig.data[0]["name"] == "Class 1 (Predictions): AUC=1.00"
        assert fig.data[1]["name"] == "Random decision: AUC=0.50"



"""Calibration curves for classification models."""

import logging

import numpy as np
import pandas as pd
import plotly.graph_objs as go
from plotly.colors import DEFAULT_PLOTLY_COLORS
from statsmodels.stats.proportion import proportion_confint

logger = logging.getLogger(__name__)


def plot_calibration_curves(
    y_true: np.ndarray,
    y_prob: np.ndarray,
    class_names: dict = None,
    ci: float = 0.90,
    n_bins: int = 10,
):
    """Plots the calibration curve for predicted vs actual class probabilities for a
    binary or multi-class classification model.

    Args:
        y_true (array): An array containing the true outcomes.
        y_prob (array): An array containing the predicted probabilities.
        class_names (dict): A dictionary containing the name to display for each class
            (e.g. {1: "Class 1", 2: "Class 2", ...}).
        ci (float): The confidence interval to use for the calibration curve error bars.
        n_bins (int): The number of bins to build the calibration curve.

    Returns:
        go.Figure: The calibration curve.
    """
    classes = np.unique(y_true)
    if not class_names:
        class_names = {c: str(c) for c in classes}
    calibration_curves = calculate_calibration_curves(
        y_true, y_prob, class_names, ci, n_bins
    )
    plot = create_calibration_curve_plot(calibration_curves)
    layout = create_calibration_curve_layout()
    fig = go.Figure(plot, layout)
    return fig


def calculate_calibration_curves(
    y_true: np.ndarray, y_prob: np.ndarray, class_names: dict, ci: float, n_bins: int
):
    """Calculates the calibration curve for each class.

    Args:
        y_true (array): An array containing the true outcomes.
        y_prob (array): An array containing the predicted probabilities.
        class_names (dict): A dictionary containing the name to display for each class
            (e.g. {1: "Class 1", 2: "Class 2", ...}).
        ci (float): The confidence interval to use for the calibration curve error bars.
        n_bins (int): The number of bins to build the calibration curve.

    Returns:
        dict: A dictionary containing the calibration curve for each class.
    """
    df = pd.DataFrame({"y_true": y_true, "y_prob": [x for x in y_prob]})
    cal_curve = {}
    for i, c in enumerate(class_names):
        proba = [x[i] for x in y_prob]
        bins = np.linspace(0, 1, num=n_bins + 1)
        proba_bins = pd.cut(
            proba, bins=bins, labels=(bins[:-1] + bins[1:]) / 2, include_lowest=True
        ).astype(float)
        df["is_true_class"] = (df["y_true"] == c).astype(float)
        avg_class = df.groupby(proba_bins)["is_true_class"].mean()
        correct = df.groupby(proba_bins)["is_true_class"].sum()
        counts = df.groupby(proba_bins).size()
        ci_plus, ci_minus = proportion_confint(
            count=correct, nobs=counts, alpha=1 - ci, method="wilson"
        )
        cal_curve[class_names[c]] = {
            "predicted": avg_class.index,
            "actual": avg_class,
            "counts": counts.tolist(),
            "error_plus": (ci_plus - avg_class),
            "error_minus": (avg_class - ci_minus),
        }
    return cal_curve


def create_calibration_curve_plot(calibration_curves):
    """Creates the data for the calibration curves plot.

    :param dict calibration_curves:
        A dictionary containing the calibration curves to plot.

    :returns list:
        A list of plotly traces containing the calibration curves.
    """
    plots = [
        go.Scatter(
            x=[0, 1],
            y=[0, 1],
            mode="lines",
            line={"dash": "dash", "color": "black"},
            showlegend=False,
        )
    ]
    for i, c in enumerate(calibration_curves):
        color = DEFAULT_PLOTLY_COLORS[i]
        plots.append(
            go.Scatter(
                x=calibration_curves[c]["predicted"],
                y=calibration_curves[c]["actual"],
                mode="markers",
                error_y={
                    "type": "data",
                    "array": calibration_curves[c]["error_plus"],
                    "arrayminus": calibration_curves[c]["error_minus"],
                },
                marker_color=color,
                legendgroup=c,
                legendgrouptitle={"text": c},
                name="Probabilities",
            )
        )
        plots.append(
            go.Bar(
                x=calibration_curves[c]["predicted"],
                y=calibration_curves[c]["counts"],
                yaxis="y2",
                opacity=0.2,
                marker_color=color,
                legendgroup=c,
                name="Number of outcomes",
            )
        )
    return plots


def create_calibration_curve_layout():
    """Creates the layout for the calibration curves plot.

    :returns plotly.graph_objs.Layout:
        The layout for the plotly calibration curves plot.
    """
    layout = go.Layout(
        xaxis={"title": "Predicted probability", "range": [0, 1]},
        yaxis={"title": "Actual probability", "range": [0, 1]},
        yaxis2={
            "title": "Number of outcomes",
            "overlaying": "y",
            "side": "right",
            "showgrid": False,
        },
        legend={"traceorder": "grouped", "orientation": "h", "y": -0.1},
        width=800,
        height=800,
    )
    return layout



"""Learning curves for classification and regression models."""

import logging

import numpy as np
import plotly.graph_objs as go
from plotly.colors import DEFAULT_PLOTLY_COLORS
from sklearn.model_selection import learning_curve
from statsmodels.stats.proportion import proportion_confint

logger = logging.getLogger(__name__)


def plot_learning_curves(
    estimator,
    X,
    y,
    scoring,
    nb_points=10,
    cv=5,
    ci=0.90,
    n_jobs=-1,
):
    """Displays the learning curve for the training data and the cross-validation data
    as a function of the number of samples in the training data.

    :type estimator: A sklearn estimator.
    :param estimator:
        The estimator for which to plot the learning curve.
    :param pandas.DataFrame X:
        A pandas DataFrame containing the features for training the estimator.
    :param pandas.Series y:
        A pandas Series containing the training target variable on which
        to train the estimator.
    :param str scoring:
        A sklearn keyword for the metric used to evaluate the estimator performance.
    :param int nb_points:
        The number of points in the learning curve.
    :param int cv:
        The number of cross-validation folds to use.
    :param int n_jobs:
        The number of parallel jobs to use to evaluate the cross validation score.
    """
    train_sizes, train_scores, val_scores = learning_curve(
        estimator,
        X,
        y,
        cv=cv,
        n_jobs=n_jobs,
        scoring=scoring,
        shuffle=True,
        train_sizes=np.linspace(1 / nb_points, 1.0, nb_points),
    )
    train_scores, val_scores = train_scores.mean(axis=1), val_scores.mean(axis=1)
    plot = create_learning_curves_plot_data(train_sizes, train_scores, val_scores, ci)
    layout = create_learning_curves_plot_layout(scoring)
    fig = go.Figure(plot, layout)
    return fig


def create_learning_curves_plot(train_sizes, train_scores, val_scores, scoring):
    """Generates a plotly figure for the learning curve.

    :param array train_sizes:
        An array containing the training sizes used to generate the learning curve.
    :param array train_scores:
        A 2D array containing the training scores (one per cross-validation fold)
        for each training size.
    :param array val_scores:
        A 2D array containing the validation scores (one per cross-validation fold)
        for each training size.
    :param str scoring:
        A sklearn keyword for the metric used to evaluate the estimator performance.

    :returns plotly.graph_objs.Figure:
        The figure containing the learning curves.
    """
    data = create_learning_curves_plot_data(train_sizes, train_scores, val_scores)
    layout = create_learning_curves_plot_layout(scoring)
    fig = go.Figure(data=data, layout=layout)
    return fig


def create_learning_curves_plot_data(train_sizes, train_scores, val_scores, ci):
    """Creates the data for the plotly learning curve plot.

    :param array train_sizes:
        An array containing the training sizes used to generate the learning curve.
    :param array train_scores:
        A 2D array containing the training scores (one per cross-validation fold)
        for each training size.
    :param array val_scores:
        A 2D array containing the validation scores (one per cross-validation fold)
        for each training size.

    :returns list:
        A list of plotly traces containing the learning curve.
    """
    train_ci_lower, train_ci_upper = get_wilson_ci(train_sizes, train_scores, ci)
    val_ci_lower, val_ci_upper = get_wilson_ci(train_sizes, val_scores, ci)
    plot_data = [
        go.Scatter(
            x=train_sizes,
            y=train_scores,
            error_y={
                "type": "data",
                "array": train_ci_upper,
                "arrayminus": train_ci_lower,
            },
            name="Training score",
            mode="lines+markers",
            marker={"color": DEFAULT_PLOTLY_COLORS[0]},
        ),
        go.Scatter(
            x=train_sizes,
            y=val_scores,
            error_y={"type": "data", "array": val_ci_upper, "arrayminus": val_ci_lower},
            name="Validation score",
            mode="lines+markers",
            marker={"color": DEFAULT_PLOTLY_COLORS[1]},
        ),
    ]
    return plot_data


def get_wilson_ci(sizes, scores, ci):
    """Returns the Wilson confidence interval for a given proportion.

    :returns float:
        The Wilson confidence interval.
    """
    all_negative = False
    if all(scores < 0):
        scores = -scores
        all_negative = True
    ci = [
        proportion_confint(count=score * size, nobs=size, alpha=1 - ci, method="wilson")
        for size, score in zip(sizes, scores)
    ]
    ci_lower_diff = scores - np.array(ci)[:, 0]
    ci_upper_diff = np.array(ci)[:, 1] - scores
    if all_negative:
        ci_lower_diff = -ci_lower_diff
        ci_upper_diff = -ci_upper_diff
    return ci_lower_diff, ci_upper_diff


def create_learning_curves_plot_layout(scoring):
    """Creates the layout for the plotly learning curve plot.

    :param str scoring:
        The name of the metrics for generating the learning curve.

    :returns plotly.graph_objs.Layout:
        The layout for the plotly learning curve plot.
    """
    layout = go.Layout(
        width=800,
        height=600,
        xaxis={"title_text": "Training set size"},
        yaxis={"title_text": f"Score ({scoring})"},
        legend={"orientation": "h", "y": -0.1},
    )
    return layout



"""Functions for plotting confusion matrix for classification models."""

import numpy as np
import plotly.graph_objs as go
from plotly.colors import DEFAULT_PLOTLY_COLORS
from sklearn.metrics import confusion_matrix


def plot_confusion_matrix(y_true, y_pred, class_names):
    """Plots the confusion matrix for a multi class clasification model.
    Normalization can be applied by setting `normalize=True`.

    :param pandas.Series y_true:
        A pandas Series containing the true target values.
    :param pandas.Series y_pred:
        A pandas Series containing the predicted target values.
    :param dict class_names:
        A dictionary containing the name to display for each class
        (e.g. {1: "Class 1", 2: "Class 2", ...}).

    :returns plotly.graph_objs.Figure:
        The figure containing the confusion matrix.
    """
    cm = calculate_confusion_matrix(y_true, y_pred, class_names)
    plot_data = create_plot_data(cm, class_names)
    layout = create_plot_layout(cm, class_names)
    fig = go.Figure(data=plot_data, layout=layout)
    return fig


def calculate_confusion_matrix(y_true, y_pred, class_names):
    """Creates the confusion matrix based on the actual and predicted values for
    each class.

    :param pandas.Series y_true:
        A pandas Series containing the true target values.
    :param pandas.Series y_pred:
        A pandas Series containing the predicted target values.
    :param dict class_names:
        A dictionary containing the name to display for each class
        (e.g. {1: "Class 1", 2: "Class 2", ...}).

    :returns array:
        The confusion matrix.
    """
    class_values = list(class_names.keys())
    cm = confusion_matrix(y_true, y_pred, labels=class_values)
    cm = cm[::-1]
    return cm


def create_plot_data(cm, class_names):
    """Creates the plot data for the confusion matrix.

    :param array cm:
        The confusion matrix.
    :param dict class_names:
        A dictionary containing the name to display for each class
        (e.g. {1: "Class 1", 2: "Class 2", ...}).

    :returns go.Heatmap:
        The plot data for the plotly confusion matrix.
    """
    labels = list(class_names.values())
    return [
        go.Heatmap(
            z=np.identity(len(class_names))[::-1],
            y=labels[::-1],
            x=labels,
            text=cm,
            colorscale=[[0, DEFAULT_PLOTLY_COLORS[3]], [1, DEFAULT_PLOTLY_COLORS[2]]],
            opacity=0.4,
            showscale=False,
            xgap=5,
            ygap=5,
        )
    ]


def create_plot_layout(cm, class_names):
    """Creates the layout for the confusion matrix.

    :param array cm:
        The confusion matrix.
    :param dict class_names:
        A dictionary containing the name to display for each class
        (e.g. {1: "Class 1", 2: "Class 2", ...}).

    :returns plotly.graph_objs.Layout:
        The layout for the confusion matrix plot.
    """
    annotations = create_confusion_matrix_annotations(cm, class_names)
    layout = go.Layout(
        xaxis={
            "title": "Predicted Outcome",
            "side": "top",
            "domain": (0.1, 1.0),
            "showgrid": False,
        },
        yaxis={"title": "Actual Outcome", "showgrid": False},
        annotations=annotations,
        height=600,
        width=800,
    )
    return layout


def create_confusion_matrix_annotations(cm, class_names):
    """Creates the annotations for the plotly confusion matrix plot.
    The value of the confusion matrix at each point is annotated onto the plot.

    :param array cm:
        The confusion matrix.
    :param dict class_names:
        A dictionary containing the name to display for each class
        (e.g. {1: "Class 1", 2: "Class 2", ...}).

    :returns list:
        A list of dictionaries specifiying the annotations for
        the confusion matrix plot.
    """
    labels = list(class_names.values())
    annotations = []
    total = np.sum(cm, axis=None)
    for i, row in enumerate(cm):
        for j, value in enumerate(row):
            annotations.append(
                {
                    "x": labels[j],
                    "y": labels[::-1][i],
                    "font": {"color": "black"},
                    "text": "<b> {:d}  <br> ({:.1%}) </b>".format(value, value / total),
                    "showarrow": False,
                }
            )
    return annotations



import logging

import numpy as np
import plotly.graph_objs as go
from sklearn.model_selection import cross_val_score
from tqdm.auto import tqdm

logger = logging.getLogger(__name__)


def plot_feature_importance(
    estimator, X, y=None, importance_type=None, scoring=None, cv=5, n_jobs=-1, max_nb=20
):
    """Displays the feature importance for the estimator, using the default feature
    importance for tree-based models or the removal importance (obtained by training
    the model after removing each feature calculated using cross-validation).

    Args:
        estimator (BaseEstimator): A fitted estimator that accepts the selected method.
        X (pd.DataFrame): The features for which to display the importance.
        y (pd.Series): The target variable (for removal importance)
        importance_type (str | None): The type of feature importance to display:
            * None: default feature importance for tree-based models
            * 'removal': removal importance
        scoring (str): The name of the  scoring function for which to estimate the
            importance of each feature.
        cv (int): The number of cross-validation folds to use.
        n_jobs (int): The number of parallel jobs to use to evaluate cross-validation
            scores.
        max_nb (int): The maximum number of features  importance to display.

    Returns:
        go.Figure: The feature importance plot.
    """
    importance = get_feature_importance(
        estimator,
        X,
        y,
        importance_type=importance_type,
        scoring=scoring,
        cv=cv,
        n_jobs=n_jobs,
    )
    data = feature_importance_plot_data(importance, max_nb)
    layout = feature_importance_layout(importance_type, scoring)
    return go.Figure(data, layout)


def get_feature_importance(estimator, X, y, importance_type, scoring, cv=5, n_jobs=1):
    """Returns the feature importance for the estimator, using the default feature
    importance for tree-based models or the removal importance (obtained by training
    the model after removing each feature calculated using cross-validation).

    Args:
        estimator (BaseEstimator): A fitted estimator that accepts the selected method.
        X (pd.DataFrame): The features for which to display the importance.
        y (pd.Series): The target variable (for removal importance)
        importance_type (str | None): The type of feature importance to display:
            * None: default feature importance for tree-based models
            * 'removal': removal importance
        scoring (str): The name of the  scoring function for which to estimate the
            importance of each feature.
        cv (int): The number of cross-validation folds to use.
        n_jobs (int): The number of parallel jobs to use to evaluate cross-validation
            scores.

    Returns:
        dict: The importance of each feature in descending order.
    """
    if importance_type is None:
        importance = tree_model_importance(estimator, X)
    elif importance_type == "removal":
        importance = removal_importance(estimator, X, y, scoring, cv, n_jobs)
    importance = dict(sorted(importance.items(), key=lambda t: t[1], reverse=True))
    return importance


def tree_model_importance(estimator, X):
    """Returns the default feature importance of a tree-based estimator. An
    exception is raised if the estimator does not have a 'feature_importances_'
    attribute.

    :type estimator: A fitted estimator that accepts the selected method.
    :param estimator:
        The fitted estimator for which to estimate the importance of each feature.
    :param pandas.DataFrame X:
        A pandas DataFrame containing the features for which to display the
        importance.

    :returns dict:
        A dictionary containing the default feature importance of a tree-based
        estimator.
    """
    if hasattr(estimator, "feature_importances_"):
        return {f: i for f, i in zip(X.columns, estimator.feature_importances_)}
    raise ValueError(
        "The estimator is not a tree-based model. "
        "Use 'removal' importance type instead."
    )


def removal_importance(
    estimator,
    X,
    y,
    scoring,
    cv=5,
    n_jobs=-1,
):
    """Returns the removal importance of each feature for the estimator using
    cross-validation. The removal importance corresponds to the decreases
    in the estimator score when the feature is removed.

    :type estimator: A fitted estimator that accepts the selected method.
    :param estimator:
        The fitted estimator for which to estimate the importance of each feature.
    :param pandas.DataFrame X:
        A pandas DataFrame containing the features for which to calculate the
        importance.
    :param pandas.Series y:
        A pandas Series containing the training target variable to calculate the
        estimator's removal importance.
    :type scorer: A function that takes y_true and y_pred as arguments.
    :param scorer:
        The scorer for which to estimate the importance of each feature. Extra
        arguments for the scorer can also be passed (e.g. multi_class='ovr')
    :param str method:
        The name of the estimator method used to make predictions ('predict' or
        'predict_proba').
    :param bool greater_is_better:
        A flag to indicate that a higher score is better. Must be set to false when
        the scorer used is an error or loss function (e.g. mean_absolute_error)
    :param int cv:
        The number of cross-validation folds to use.
    :param int n_jobs:
        The number of parallel jobs to use to evaluate the cross validation score.

    :returns dict:
        A dictionary containing the removal importance for each feature.
    """
    importance = {col: [] for col in X.columns}
    base_score = cross_val_score(
        estimator, X, y, scoring=scoring, cv=cv, n_jobs=n_jobs
    ).mean()
    X = X.copy()
    for col in tqdm(X.columns, desc="Calculating feature importance"):
        old_X_col = X[col].copy()
        X[col] = 0
        new_score = cross_val_score(
            estimator, X, y, scoring=scoring, cv=cv, n_jobs=n_jobs
        ).mean()
        X[col] = old_X_col.copy()
        importance[col] = base_score - new_score
    return importance


def feature_importance_plot_data(importance, max_nb):
    """Returns a bar plot of feature importance.

    :param collections.OrderedDict importance:
        A dictionary containing the importance of each feature in descending order.
    :param int max_nb:
        The maximum number of features for which to display the importance.

    :return go.Figure:
        A figure containing a bar plot of feature importance.
    """
    features = [f.replace("_", " ").capitalize() for f in importance][:max_nb]
    importance_vals = list(importance.values())[:max_nb]
    min_scale = np.min(importance_vals) - np.max(importance_vals)
    plot_data = go.Bar(
        x=importance_vals,
        y=features,
        orientation="h",
        text=features,
        textposition="auto",
        marker={"color": importance_vals, "colorscale": "Blues", "cmid": min_scale},
        textfont_size=12,
    )
    return plot_data


def feature_importance_layout(importance_type, scorer):
    x_label = "Feature importance"
    if importance_type == "removal":
        x_label += f" (impact on {scorer} score)"
    return go.Layout(
        yaxis={"autorange": "reversed", "tickvals": []},
        xaxis={"title": x_label},
        height=700,
        width=1000,
    )
